//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

/**
 * @brief Allow socket to download limit number of bytes
 *
 * This function is called every 0.1 second by CPartFile::ProcessDL()
 * @param limit Number of bytes to send during this 0.1 second slot 
 * @return void
 */
void CEMSocket::SetDownloadLimit(uint32 limit){	
	// MOD BEGIN netfinity: Accumulate download limits
	downloadLimit += limit;	
	downloadLimitEnable = true;	
	if(downloadLimit > 20 * limit && downloadLimit > 4500) // Allow a maximum of 2.0 sec to accumulate or 3 * MTU
		downloadLimit = max(20 * limit, 4500); 
	// MOD BEGIN netfinity

	// CPU load improvement
	if(limit > 0 && pendingOnReceive == true){
		OnReceive(0);
	}
}

/**
 * @brief Allow socket to download as much data as possible
 *
 * @return void
 */
void CEMSocket::DisableDownloadLimit(){
	downloadLimitEnable = false;

	// CPU load improvement
	if(pendingOnReceive == true){
		OnReceive(0);
	}
}
